"""Command implementation."""


class Command(object):
    """The object for coap commands."""

    def __init__(self, method, path, data=None, *, parse_json=True, timeout=10,
                 observe=False, observe_duration=0, process_result=None,
                 err_callback=None):
        self._method = method
        self._path = path
        self._data = data
        self._parse_json = parse_json
        self._timeout = timeout
        self._process_result = process_result
        self._err_callback = err_callback
        self._observe = observe
        self._observe_duration = observe_duration
        self._raw_result = None
        self._result = None

    @property
    def method(self):
        return self._method

    @property
    def path(self):
        return self._path

    @property
    def data(self):
        return self._data

    @property
    def parse_json(self):
        return self._parse_json

    @property
    def timeout(self):
        return self._timeout

    @property
    def process_result(self):
        return self._process_result

    @property
    def err_callback(self):
        """This will be fired when an observe request fails."""
        return self._err_callback

    @property
    def observe(self):
        return self._observe

    @property
    def observe_duration(self):
        return self._observe_duration

    @property
    def raw_result(self):
        return self._raw_result

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, value):
        """The result of the command."""
        if self._process_result:
            self._result = self._process_result(value)

        self._raw_result = value

    def url(self, host):
        """Generate url for coap client."""
        path = '/'.join(str(v) for v in self._path)
        return 'coaps://{}:5684/{}'.format(host, path)



"""Classes to interact with devices."""
from datetime import datetime


from .command import Command
from .const import (
    ROOT_DEVICES,
    ATTR_APPLICATION_TYPE,
    ATTR_DEVICE_INFO,
    ATTR_REACHABLE_STATE,
    ATTR_LAST_SEEN,
    ATTR_LIGHT_CONTROL,
    ATTR_LIGHT_STATE,
    ATTR_LIGHT_DIMMER,
    ATTR_LIGHT_COLOR_X,
    ATTR_LIGHT_COLOR_Y,
    ATTR_LIGHT_COLOR,
    ATTR_TRANSITION_TIME
)
from .color import can_kelvin_to_xy, kelvin_to_xyY, xyY_to_kelvin, rgb_to_xyY,\
    xy_brightness_to_rgb, COLORS, MIN_KELVIN, MAX_KELVIN,\
    MIN_KELVIN_WS, MAX_KELVIN_WS
from .resource import ApiResource
from .error import ColorError


class Device(ApiResource):
    """Base class for devices."""

    @property
    def application_type(self):
        return self.raw.get(ATTR_APPLICATION_TYPE)

    @property
    def path(self):
        return [ROOT_DEVICES, self.id]

    @property
    def device_info(self):
        return DeviceInfo(self)

    @property
    def last_seen(self):
        if ATTR_LAST_SEEN not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_LAST_SEEN])

    @property
    def reachable(self):
        return self.raw.get(ATTR_REACHABLE_STATE) == 1

    @property
    def has_light_control(self):
        return (self.raw is not None and
                len(self.raw.get(ATTR_LIGHT_CONTROL, "")) > 0)

    @property
    def light_control(self):
        return LightControl(self)

    def __repr__(self):
        return "<{} - {} ({})>".format(self.id, self.name,
                                       self.device_info.model_number)


class DeviceInfo:
    """Represent device information.

    http://www.openmobilealliance.org/tech/profiles/LWM2M_Device-v1_0.xml
    """
    ATTR_MANUFACTURER = "0"
    ATTR_MODEL_NUMBER = "1"
    ATTR_SERIAL = "2"
    ATTR_FIRMWARE_VERSION = "3"
    ATTR_POWER_SOURCE = "6"
    VALUE_POWER_SOURCES = {
        1: 'Internal Battery',
        2: 'External Battery',
        3: 'Battery',  # Not in spec, used by remote
        4: 'Power over Ethernet',
        5: 'USB',
        6: 'AC (Mains) power',
        7: 'Solar'
    }
    ATTR_BATTERY = "9"

    def __init__(self, device):
        self._device = device

    @property
    def manufacturer(self):
        """Human readable manufacturer name."""
        return self.raw.get(DeviceInfo.ATTR_MANUFACTURER)

    @property
    def model_number(self):
        """A model identifier string (manufactuer specified string)."""
        return self.raw.get(DeviceInfo.ATTR_MODEL_NUMBER)

    @property
    def serial(self):
        """Serial number string."""
        return self.raw.get(DeviceInfo.ATTR_SERIAL)

    @property
    def firmware_version(self):
        """Current firmware version string of the device."""
        return self.raw.get(DeviceInfo.ATTR_FIRMWARE_VERSION)

    @property
    def power_source(self):
        """Power source."""
        return self.raw.get(DeviceInfo.ATTR_POWER_SOURCE)

    @property
    def power_source_str(self):
        """String representation of current power source."""
        if DeviceInfo.ATTR_POWER_SOURCE not in self.raw:
            return None
        return DeviceInfo.VALUE_POWER_SOURCES.get(self.power_source, 'Unknown')

    @property
    def battery_level(self):
        """Battery in 0..100"""
        return self.raw.get(DeviceInfo.ATTR_BATTERY)

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ATTR_DEVICE_INFO]


class LightControl:
    """Class to control the lights."""

    def __init__(self, device):
        self._device = device

    @property
    def lights(self):
        """Return light objects of the light control."""
        return [Light(self._device, i) for i in range(len(self.raw))]

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ATTR_LIGHT_CONTROL]

    @property
    def can_set_kelvin(self):
        """Return whether controlled light supports color temperature.
        The range of supported tempertures is defined by properties
        min_kelvin and max_kelvin."""
        return 'WS' in self._device.device_info.model_number

    @property
    def can_set_color(self):
        """Return whether controlled light supports arbitrary color."""
        return 'CWS' in self._device.device_info.model_number

    @property
    def _kelvin_range(self):
        if not self.can_set_kelvin:
            # White bulb
            return (None, None)
        if not self.can_set_color:
            # White spectrum bulb
            return (MIN_KELVIN_WS, MAX_KELVIN_WS)
        # Color bulb
        return (MIN_KELVIN, MAX_KELVIN)

    @property
    def min_kelvin(self):
        """Return minimum supported color temperature."""
        return self._kelvin_range[0]

    @property
    def max_kelvin(self):
        """Return maximum supported color temperature."""
        return self._kelvin_range[1]

    def set_state(self, state, *, index=0):
        """Set state of a light."""
        return self.set_values({
            ATTR_LIGHT_STATE: int(state)
        }, index=index)

    def set_dimmer(self, dimmer, *, index=0, transition_time=None):
        """Set dimmer value of a light.

        dimmer: Integer between 0..255
        transition_time: Integer representing tenth of a second (default None)
        """
        values = {
            ATTR_LIGHT_DIMMER: dimmer
        }
        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time

        return self.set_values(values, index=index)

    def set_hex_color(self, color, *, index=0):
        """Set xy color of the light."""
        return self.set_values({
            ATTR_LIGHT_COLOR: color,
        }, index=index)

    def set_xy_color(self, color_x, color_y, *, index=0):
        """Set xy color of the light."""
        return self.set_values({
            ATTR_LIGHT_COLOR_X: color_x,
            ATTR_LIGHT_COLOR_Y: color_y
        }, index=index)

    def set_kelvin_color(self, kelvins, *, index=0):
        return self.set_values(kelvin_to_xyY(kelvins), index=index)

    def set_predefined_color(self, colorname, *, index=0):
        try:
            color = COLORS[colorname.lower().replace(" ", "_")]
            return self.set_hex_color(color, index=index)
        except KeyError:
            raise ColorError('Invalid color specified: %s',
                             colorname)

    def set_rgb_color(self, r, g, b, *, index=0):
        return self.set_values(rgb_to_xyY(r, g, b), index=index)

    def set_values(self, values, *, index=0):
        """
        Set values on light control.

        Returns a Command.
        """
        assert len(self.raw) == 1, \
            'Only devices with 1 light supported'

        return Command('put', self._device.path, {
            ATTR_LIGHT_CONTROL: [
                values
            ]
        })

    def __repr__(self):
        return '<LightControl for {} ({} lights)>'.format(self._device.name,
                                                          len(self.raw))


class Light:
    """Represent a light control.

    https://github.com/IPSO-Alliance/pub/blob/master/docs/IPSO-Smart-Objects.pdf
    """

    def __init__(self, device, index):
        self.device = device
        self.index = index

    @property
    def state(self):
        return self.raw.get(ATTR_LIGHT_STATE) == 1

    @property
    def dimmer(self):
        return self.raw.get(ATTR_LIGHT_DIMMER)

    @property
    def hex_color(self):
        return self.raw.get(ATTR_LIGHT_COLOR)

    @property
    def hex_color_inferred(self):
        raw_color = self.hex_color
        if raw_color is not None and len(raw_color) == 6:
            return raw_color
        (x, y) = self.xy_color_inferred

        def scale(val):
            return float(val)/65535
        return '{0:02x}{1:02x}{2:02x}'.format(
                *xy_brightness_to_rgb(scale(x), scale(y), self.dimmer)
        )

    @property
    def xy_color(self):
        return (self.raw.get(ATTR_LIGHT_COLOR_X),
                self.raw.get(ATTR_LIGHT_COLOR_Y))

    @property
    def xy_color_inferred(self):
        (current_x, current_y) = self.xy_color
        if current_x is not None and current_y is not None:
            return (self.raw.get(ATTR_LIGHT_COLOR_X),
                    self.raw.get(ATTR_LIGHT_COLOR_Y))
        # White Tradfri has no x and y, but spec provide 2700K value
        xy = kelvin_to_xyY(2700)
        return (xy[ATTR_LIGHT_COLOR_X], xy[ATTR_LIGHT_COLOR_Y])

    @property
    def kelvin_color_inferred(self):
        current_x = self.raw.get(ATTR_LIGHT_COLOR_X)
        current_y = self.raw.get(ATTR_LIGHT_COLOR_Y)
        if current_x is not None and current_y is not None:
            kelvin = xyY_to_kelvin(current_x, current_y)
            # Only return a kelvin value if it is inside the range that the
            # kelvin->xyY function supports
            return kelvin if can_kelvin_to_xy(kelvin) else None
        # White Tradfri has no x and y, but spec provide 2700K value
        return 2700

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self.device.raw[ATTR_LIGHT_CONTROL][self.index]

    def __repr__(self):
        state = "on" if self.state else "off"
        return "<Light #{} - " \
               "name: {}, " \
               "state: {}, " \
               "dimmer: {}, "\
               "hex_color: {}, " \
               "xy_color: {}, " \
               ">".format(self.index, self.device.name, state, self.dimmer,
                          self.hex_color, self.xy_color)



"""Coap implementation using aiocoap."""
import asyncio
import json
import logging

from aiocoap import Message, Context
from aiocoap.error import RequestTimedOut, Error, ConstructionRenderableError
from aiocoap.numbers.codes import Code
from aiocoap.transports import tinydtls

from ..error import ClientError, ServerError, RequestTimeout
from ..gateway import Gateway

_LOGGER = logging.getLogger(__name__)


class PatchedDTLSSecurityStore:
    """Patched DTLS store in lieu of impl."""

    IDENTITY = None
    KEY = None

    def _get_psk(self, host, port):
        return PatchedDTLSSecurityStore.IDENTITY, PatchedDTLSSecurityStore.KEY


tinydtls.DTLSSecurityStore = PatchedDTLSSecurityStore


class APIFactory:
    def __init__(self, host, psk_id='pytradfri', psk=None, loop=None):
        self._psk = psk
        self._host = host
        self._psk_id = psk_id
        self._loop = loop
        self._observations_err_callbacks = []
        self._protocol = None

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        PatchedDTLSSecurityStore.IDENTITY = self._psk_id.encode('utf-8')

        if self._psk:
            PatchedDTLSSecurityStore.KEY = self._psk.encode('utf-8')

    @property
    def psk_id(self):
        return self._psk_id

    @psk_id.setter
    def psk_id(self, value):
        self._psk_id = value
        PatchedDTLSSecurityStore.IDENTITY = self._psk_id.encode('utf-8')

    @property
    def psk(self):
        return self._psk

    @psk.setter
    def psk(self, value):
        self._psk = value
        PatchedDTLSSecurityStore.KEY = self._psk.encode('utf-8')

    @asyncio.coroutine
    def _get_protocol(self):
        """Get the protocol for the request."""
        if not self._protocol:
            self._protocol = yield from Context.create_client_context(
                loop=self._loop)
        return self._protocol

    @asyncio.coroutine
    def _reset_protocol(self, exc=None):
        """Reset the protocol if an error occurs.
           This can be removed when chrysn/aiocoap#79 is closed."""
        # Be responsible and clean up.
        protocol = yield from self._get_protocol()
        yield from protocol.shutdown()
        self._protocol = None
        # Let any observers know the protocol has been shutdown.
        for ob_error in self._observations_err_callbacks:
            ob_error(exc)
        self._observations_err_callbacks.clear()

    @asyncio.coroutine
    def _get_response(self, msg):
        """Perform the request, get the response."""
        try:
            protocol = yield from self._get_protocol()
            pr = protocol.request(msg)
            r = yield from pr.response
            return pr, r
        except ConstructionRenderableError as e:
            raise ClientError("There was an error with the request.", e)
        except RequestTimedOut as e:
            yield from self._reset_protocol(e)
            raise RequestTimeout('Request timed out.', e)
        except Error as e:
            yield from self._reset_protocol(e)
            raise ServerError("There was an error with the request.", e)

    @asyncio.coroutine
    def _execute(self, api_command):
        """Execute the command."""
        if api_command.observe:
            yield from self._observe(api_command)
            return

        method = api_command.method
        path = api_command.path
        data = api_command.data
        parse_json = api_command.parse_json
        url = api_command.url(self._host)

        kwargs = {}

        if data is not None:
            kwargs['payload'] = json.dumps(data).encode('utf-8')
            _LOGGER.debug('Executing %s %s %s: %s', self._host, method, path,
                          data)
        else:
            _LOGGER.debug('Executing %s %s %s', self._host, method, path)

        api_method = Code.GET
        if method == 'put':
            api_method = Code.PUT
        elif method == 'post':
            api_method = Code.POST
        elif method == 'delete':
            api_method = Code.DELETE
        elif method == 'fetch':
            api_method = Code.FETCH
        elif method == 'patch':
            api_method = Code.PATCH

        msg = Message(code=api_method, uri=url, **kwargs)

        _, res = yield from self._get_response(msg)

        api_command.result = _process_output(res, parse_json)

        return api_command.result

    @asyncio.coroutine
    def request(self, api_commands):
        """Make a request."""
        if not isinstance(api_commands, list):
            result = yield from self._execute(api_commands)
            return result

        commands = (self._execute(api_command) for api_command in api_commands)
        command_results = yield from asyncio.gather(*commands, loop=self._loop)

        return command_results

    @asyncio.coroutine
    def _observe(self, api_command):
        """Observe an endpoint."""
        duration = api_command.observe_duration
        url = api_command.url(self._host)
        err_callback = api_command.err_callback

        msg = Message(code=Code.GET, uri=url, observe=duration)

        # Note that this is necessary to start observing
        pr, r = yield from self._get_response(msg)

        api_command.result = _process_output(r)

        def success_callback(res):
            api_command.result = _process_output(res)

        def error_callback(ex):
            err_callback(ex)

        ob = pr.observation
        ob.register_callback(success_callback)
        ob.register_errback(error_callback)
        self._observations_err_callbacks.append(ob.error)

    @asyncio.coroutine
    def generate_psk(self, security_key):
        """
        Generate and set a psk from the security key.
        """
        if not self._psk:
            PatchedDTLSSecurityStore.IDENTITY = 'Client_identity'.encode(
                'utf-8')
            PatchedDTLSSecurityStore.KEY = security_key.encode('utf-8')

            command = Gateway().generate_psk(self._psk_id)
            self._psk = yield from self.request(command)

            PatchedDTLSSecurityStore.IDENTITY = self._psk_id.encode('utf-8')
            PatchedDTLSSecurityStore.KEY = self._psk.encode('utf-8')

            # aiocoap has now cached our psk, so it must be reset.
            # We also no longer need the protocol, so this will shutdown that.
            yield from self._reset_protocol()

        return self._psk


def _process_output(res, parse_json=True):
    """Process output."""
    res_payload = res.payload.decode('utf-8')
    output = res_payload.strip()

    _LOGGER.debug('Status: %s, Received: %s', res.code, output)

    if not output:
        return None

    if not res.code.is_successful:
        if res.code >= 128 and res.code < 160:
            raise ClientError(output)
        elif res.code >= 160 and res.code < 192:
            raise ServerError(output)

    if not parse_json:
        return output

    return json.loads(output)



from .const import (ATTR_LIGHT_COLOR_X as X, ATTR_LIGHT_COLOR_Y as Y)


# Kelvin range for which the conversion functions work
# and that RGB bulbs can show
MIN_KELVIN = 1667
MAX_KELVIN = 25000

# Kelvin range that white-spectrum bulbs can actually show
MIN_KELVIN_WS = 2200
MAX_KELVIN_WS = 4000

# Extracted from Tradfri Android App string.xml
COLOR_NAMES = {
    '4a418a': 'Blue',
    '6c83ba': 'Light Blue',
    '8f2686': 'Saturated Purple',
    'a9d62b': 'Lime',
    'c984bb': 'Light Purple',
    'd6e44b': 'Yellow',
    'd9337c': 'Saturated Pink',
    'da5d41': 'Dark Peach',
    'dc4b31': 'Saturated Red',
    'dcf0f8': 'Cold sky',
    'e491af': 'Pink',
    'e57345': 'Peach',
    'e78834': 'Warm Amber',
    'e8bedd': 'Light Pink',
    'eaf6fb': 'Cool daylight',
    'ebb63e': 'Candlelight',
    'efd275': 'Warm glow',
    'f1e0b5': 'Warm white',
    'f2eccf': 'Sunrise',
    'f5faf6': 'Cool white'
}
# When setting colors by name via the API,
# lowercase strings with no spaces are preferred
COLORS = {name.lower().replace(" ", "_"): hex
          for hex, name in COLOR_NAMES.items()}


def can_kelvin_to_xy(k):
    return MIN_KELVIN <= k <= MAX_KELVIN


# Only used locally to perform normalization of x, y values
# Scaling to 65535 range and rounding
def normalize_xy(x, y):
    return (int(x*65535+0.5), int(y*65535+0.5))


def kelvin_to_xyY(T, white_spectrum_bulb=False):
    # Sources: "Design of Advanced Color - Temperature Control System
    #           for HDTV Applications" [Lee, Cho, Kim]
    # and https://en.wikipedia.org/wiki/Planckian_locus#Approximation
    # and http://fcam.garage.maemo.org/apiDocs/_color_8cpp_source.html

    # Check for Kelvin range for which this function works
    if not (MIN_KELVIN <= T <= MAX_KELVIN):
        raise ValueError('Kelvin needs to be between {} and {}'.format(
            MIN_KELVIN, MAX_KELVIN))

    # Check for White-Spectrum kelvin range
    if white_spectrum_bulb and not (MIN_KELVIN_WS <= T <= MAX_KELVIN_WS):
        raise ValueError('Kelvin needs to be between {} and {} for '
                         'white spectrum bulbs'.format(
                          MIN_KELVIN_WS, MAX_KELVIN_WS))

    if T <= 4000:
        # One number differs on Wikipedia and the paper:
        #     0.2343589 is 0.2343580 on Wikipedia... don't know why
        x = -0.2661239*(10**9)/T**3 - 0.2343589*(10**6)/T**2 \
            + 0.8776956*(10**3)/T + 0.17991
    elif T <= 25000:
        x = -3.0258469*(10**9)/T**3 + 2.1070379*(10**6)/T**2 \
            + 0.2226347*(10**3)/T + 0.24039

    if T <= 2222:
        y = -1.1063814*x**3 - 1.3481102*x**2 + 2.18555832*x - 0.20219683
    elif T <= 4000:
        y = -0.9549476*x**3 - 1.37418593*x**2 + 2.09137015*x - 0.16748867
    elif T <= 25000:
        y = 3.081758*x**3 - 5.8733867*x**2 + 3.75112997*x - 0.37001483

    x, y = normalize_xy(x, y)
    return {X: x, Y: y}


def xyY_to_kelvin(x, y):
    # This is an approximation, for information, see the source.
    # Source: https://en.wikipedia.org/wiki/Color_temperature#Approximation
    # Input range for x and y is 0-65535
    n = (x/65535-0.3320) / (y/65535-0.1858)
    kelvin = int((-449*n**3 + 3525*n**2 - 6823.3*n + 5520.33) + 0.5)
    return kelvin


def rgb2xyzA(r, g, b):
    # Uses CIE standard illuminant A = 2856K
    # src: http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    # calculation https://gist.github.com/r41d/43e14df2ccaeca56d32796efd6584b48
    X = 0.76103282*r + 0.29537849*g + 0.04208869*b
    Y = 0.39240755*r + 0.59075697*g + 0.01683548*b
    Z = 0.03567341*r + 0.0984595*g + 0.22166709*b
    return X, Y, Z


def rgb2xyzD65(r, g, b):
    # Uses CIE standard illuminant D65 = 6504K
    # src: http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    X = 0.4124564 * r + 0.3575761 * g + 0.1804375 * b
    Y = 0.2126729 * r + 0.7151522 * g + 0.0721750 * b
    Z = 0.0193339 * r + 0.1191920 * g + 0.9503041 * b
    return X, Y, Z


def xyz2xyY(X, Y, Z):
    total = X + Y + Z
    return (0, 0) if total == 0 else normalize_xy(X / total, Y / total)


def rgb_to_xyY(r, g, b):
    # According to http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html
    # and http://www.brucelindbloom.com/index.html?Eqn_XYZ_to_xyY.html
    def prepare(val):
        val = max(min(val, 255), 0) / 255.0
        if val <= 0.04045:
            return val / 12.92
        else:
            return ((val + 0.055) / 1.055) ** 2.4
    r, g, b = map(prepare, (r, g, b))

    x, y = xyz2xyY(*rgb2xyzA(r, g, b))
    return {X: x, Y: y}


# Converted to Python from Obj-C, original source from:
# http://www.developers.meethue.com/documentation/color-conversions-rgb-xy
# pylint: disable=invalid-sequence-index
def xy_brightness_to_rgb(vX: float, vY: float, ibrightness: int):
    """Convert from XYZ to RGB."""
    brightness = ibrightness / 255.
    if brightness == 0:
        return (0, 0, 0)
    Y = brightness
    if vY == 0:
        vY += 0.00000000001
    X = (Y / vY) * vX
    Z = (Y / vY) * (1 - vX - vY)
    # Convert to RGB using Wide RGB D65 conversion.
    r = X * 1.656492 - Y * 0.354851 - Z * 0.255038
    g = -X * 0.707196 + Y * 1.655397 + Z * 0.036152
    b = X * 0.051713 - Y * 0.121364 + Z * 1.011530
    # Apply reverse gamma correction.
    r, g, b = map(
        lambda x: (12.92 * x) if (x <= 0.0031308) else
        ((1.0 + 0.055) * pow(x, (1.0 / 2.4)) - 0.055),
        [r, g, b]
    )
    # Bring all negative components to zero.
    r, g, b = map(lambda x: max(0, x), [r, g, b])
    # If one component is greater than 1, weight components by that value.
    max_component = max(r, g, b)
    if max_component > 1:
        r, g, b = map(lambda x: x / max_component, [r, g, b])
    ir, ig, ib = map(lambda x: int(x * 255), [r, g, b])
    return (ir, ig, ib)



"""Implement an API wrapper around Ikea Tradfri."""

from .error import (
    PyTradFriError, RequestError, ClientError, ServerError, RequestTimeout)
from .gateway import Gateway

__all__ = ['Gateway', 'PyTradFriError', 'RequestError', 'ClientError',
           'ServerError', 'RequestTimeout']



from datetime import datetime

from .command import Command
from .const import (
    ATTR_NAME,
    ATTR_CREATED_AT,
    ATTR_ID,
)


class ApiResource:
    """Base object for resources returned from the gateway."""

    def __init__(self, raw):
        """Initialize base object."""
        self.raw = raw

    @property
    def id(self):
        return self.raw.get(ATTR_ID)

    @property
    def name(self):
        return self.raw.get(ATTR_NAME)

    @property
    def created_at(self):
        if ATTR_CREATED_AT not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_CREATED_AT])

    @property
    def path(self):
        """Path to resource."""
        raise NotImplemented('Path property needs to be implemented')

    def observe(self, callback, err_callback, duration=60):
        """Observe resource and call callback when updated."""
        def observe_callback(value):
            """
            Called when end point is updated.

            Returns a Command.
            """
            self.raw = value
            callback(self)

        return Command('get', self.path, process_result=observe_callback,
                       err_callback=err_callback,
                       observe=True,
                       observe_duration=duration)

    def set_name(self, name):
        """Set group name."""
        return self.set_values({
            ATTR_NAME: name
        })

    def set_values(self, values):
        """
        Helper to set values for group.

        Returns a Command.
        """
        return Command('put', self.path, values)

    def update(self):
        """
        Update the group.

        Returns a Command.
        """
        def process_result(result):
            self.raw = result
        return Command('get', self.path, process_result=process_result)



from setuptools import setup, find_packages

VERSION = "4.0.1"
DOWNLOAD_URL = \
    'https://github.com/ggravlingen/pytradfri/archive/{}.zip'.format(VERSION)

EXTRAS_REQUIRE = {
    'async': ['aiocoap >= 0.3', 'DTLSSocket >= 0.1.4']
}

DEP_LINKS = [
    "https://github.com/chrysn/aiocoap/archive/3286f48f0b949901c8b5c04c0719dc54ab63d431.zip#egg=aiocoap-0.3"
]

PACKAGES = find_packages(exclude=['tests', 'tests.*'])

setup(
  name='pytradfri',
  packages=PACKAGES,
  python_requires='>=3.4',
  version=VERSION,
  description='IKEA Trådfri/Tradfri API. Control and observe your '
              'lights from Python.',
  author='balloob, lwis, ggravlingen',
  author_email='no@email.com',
  url='https://github.com/ggravlingen/pytradfri',
  download_url=DOWNLOAD_URL,
  extras_require=EXTRAS_REQUIRE,
  dependency_links=DEP_LINKS,
)
